import processing.core.*; 
import processing.data.*; 
import processing.opengl.*; 

import java.applet.*; 
import java.awt.Dimension; 
import java.awt.Frame; 
import java.awt.event.MouseEvent; 
import java.awt.event.KeyEvent; 
import java.awt.event.FocusEvent; 
import java.awt.Image; 
import java.io.*; 
import java.net.*; 
import java.text.*; 
import java.util.*; 
import java.util.zip.*; 
import java.util.regex.*; 

public class KiteColoriser_v1_0 extends PApplet {

/* @pjs preload="kiteTemplate.png,GUI4Colorizer.png"; 
 */

//********************************************************************************************************************
//********************************************************************************************************************
//** ****************************************  GENERAL PURPOSE COLORIZER  ********************************************
//** 
//** Author: Narc\u00eds Par\u00e9s (nparesb at gmail dot com)
//** Copyright: This code is released under Creative Commons Attribution Non-Commercial Share Alike (CC BY-NC-SA)
//** November 2012
//** 
//** Although this colorizer has bein conceived for stunt kites, it is actually a general purpose colorizer
//** since it can be used for any shapes whatsoever as long as the original template is defined in a specific way.
//** 
//** This code is provided "as is", without warranty of any kind, express or implied, including but not limited to 
//** the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the 
//** author be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise,
//** arising from, out of or in connection with the software or the use or other dealings in the software.
//** 
//** NOTE: The images used in the colorizer are not part of the code and can be changed by the users. Hence, the author
//** of this code cannot be held liable for any missuse of copyrighted images in a specific deployment of the colorizer.
//** i.e. the images are not coded within the program or part of the final application.
//** 
//** (Please see below the code for explanations on how to prepare the template and materials)
//** 
//********************************************************************************************************************
//********************************************************************************************************************

PImage guiBase;
PImage kiteTemplate;        // The template of the kite must be kiteWidth x kiteHeight pixels 
                                       // and each independent sail layout piece must have a distinct grey scale from 1 to 254
                                       // while 0 (black) is for the frame of the kite & 255 (white) is for the background
PImage kiteShadow;          // The greyscale image used to simulate the 3D shape of the sail with a lateral lighting
PImage kiteResult;          // The resulting image of the sail pieces
boolean effect3D = true;
PImage kiteEdges;           // The edges of the kite to define clear borders or seam simulation
int numberMaterials = 22;
PImage[] materials = new PImage[24];   // Images for sail materials
int[] mappingGreys2Materials = new int[256];
int kiteWidth = 600, kiteHeight = 300;
int guiWidth = 600, guiHeight = 300, guiY;

int activeMaterial = 0;
ButtonI[] materialButtons = new ButtonI[24];

ButtonI[] backgroundButtons = new ButtonI[5];
PImage[] kiteBackground = new PImage[5];      // The Background for Kite Display
int activeBackground = 3;

ButtonC[] seamsButtons = new ButtonC[3];
int seamsState = 1;  // Whether the Panel Lines are Solid (2), Seams (1) or Hidden (0)

ButtonC resetButton;  // Resets kite layout to all white
ButtonC effect3DButton;  // Changes kite 3D effect

ButtonC saveButton;  // Saves image of kite to PNG image file format: kite.png

//*****************************************************
//* Button Classes for GUI
//* ButtonI buttons with images
//* ButtonC buttons with solid color

class ButtonI {
  int bpx, bpy, bw, bh;
  PImage bImage;
  
  ButtonI(int px, int py, int w, int h, PImage img){
    bpx = px;
    bpy = py;
    bw = w;
    bh = h;
    bImage = img;
    //dispButton();
  }

  public void dispButtonI(){
    image(bImage, bpx, bpy, bw, bh);
  }
  
  public boolean clickedButtonI(){
    return isInside(mouseX, mouseY, bpx, bpy, bw, bh);
  }

}

class ButtonC {
  int bpx, bpy, bw, bh;
  int bColor;
  
  ButtonC(int px, int py, int w, int h, int col){
    bpx = px;
    bpy = py;
    bw = w;
    bh = h;
    bColor = col;
    //dispButton();
  }

  public void setColor(int col){
      bColor = col;
  }

  public void dispButtonC(){
    noStroke();
    fill(bColor);
    rect(bpx, bpy, bw, bh);
  }
  
  public boolean clickedButtonC(){
    return isInside(mouseX, mouseY, bpx, bpy, bw, bh);
  }

}

//* end of Button Classes for GUI
//*****************************************************

//*****************************************************
//* GUI Utilities (displaying, detecting mouse clicks, etc.

public boolean isInside(int px, int py, int ax, int ay, int w, int h){
  if(px<ax) return false;
  else if(px>ax+w) return false;
  else if(py<ay) return false;
  else if(py>ay+h) return false;
  else return true;
}

//* GUI Display
public void gui(){
  stroke(180);
  fill(255);
  rect(10,10,kiteWidth,kiteHeight); // Kite
  image(guiBase, 0, guiY);
  
  // Material selectors
  for(int i=0; i<numberMaterials; i++){
    materialButtons[i].dispButtonI();
  }
  // Active Material
  image(materials[activeMaterial], 101, 218+guiY, 80, 40);
  
  // Background Selectors
  for(int i=0; i<5; i++){
    backgroundButtons[i].dispButtonI();
  }

  // Seams Choice
  for(int i=0; i<3; i++){
    seamsButtons[i].dispButtonC();
  }

  resetButton.dispButtonC();
  effect3DButton.dispButtonC();

  saveButton.dispButtonC();
}

//* General Click Control Management
public void mouseClicked(){

  // Is the click in the Kite zone?
  if(isInside(mouseX, mouseY, 10, 10, kiteWidth, kiteHeight)){
    kiteTemplate.loadPixels();
    mappingGreys2Materials[ (int)red(kiteTemplate.pixels[(mouseY-10)*kiteWidth+mouseX-10]) ] = activeMaterial;
  }
  else {
    // Is the click in the GUI zone?
    if(isInside(mouseX, mouseY, 10, kiteHeight+20, guiWidth, guiHeight)){
      // Test for Material Buttons
      for(int i=0; i<numberMaterials; i++){
        if(materialButtons[i].clickedButtonI()) activeMaterial = i;
      }

      for(int i=0; i<5; i++){
        if(backgroundButtons[i].clickedButtonI()) activeBackground = i;
      }

      if(seamsButtons[0].clickedButtonC()){  // Panel Line will be hidden
        seamsButtons[0].setColor(color(0,255,0,255));
        seamsButtons[1].setColor(color(0,255,0,0));
        seamsButtons[2].setColor(color(0,255,0,0));
        seamsState = 0;
      }
      if(seamsButtons[1].clickedButtonC()){  // Panel Lines will be seams
        seamsButtons[0].setColor(color(0,255,0,0));
        seamsButtons[1].setColor(color(0,255,0,255));
        seamsButtons[2].setColor(color(0,255,0,0));
        seamsState = 1;
      }
      if(seamsButtons[2].clickedButtonC()){  // Panel Lines will be Solid
        seamsButtons[0].setColor(color(0,255,0,0));
        seamsButtons[1].setColor(color(0,255,0,0));
        seamsButtons[2].setColor(color(0,255,0,255));
        seamsState = 2;
      }

      if(resetButton.clickedButtonC()){  // Reset of the kite layout to all white
        initMappings();
      }
      
      if(effect3DButton.clickedButtonC()){  // Toggle Effect3D and button
        if(effect3D)
          effect3DButton.setColor(color(0,0,255,0));
        else
          effect3DButton.setColor(color(0,0,255,255));
        effect3D = !effect3D;
      }

      if(saveButton.clickedButtonC()){  // Save and image of the kite to kite.png
        save("kite.png");
      }
      
    }
  }
  
}
//* end of GUI Utilities (displaying, detecting mouse clicks, etc.
//*****************************************************

//*****************************************************
//** Setup & Initialization
public void initMappings(){
   for(int i=0; i<256; i++) mappingGreys2Materials[i] = 0; 
}

public void loadMaterials(){
  materials[0] = loadImage("mat01.png");
  materials[0].loadPixels();
  materials[1] = loadImage("mat02.png");
  materials[1].loadPixels();
  materials[2] = loadImage("mat03.png");
  materials[2].loadPixels();
  materials[3] = loadImage("mat04.png");
  materials[3].loadPixels();
  materials[4] = loadImage("mat05.png");
  materials[4].loadPixels();
  materials[5] = loadImage("mat06.png");
  materials[5].loadPixels();
  materials[6] = loadImage("mat07.png");
  materials[6].loadPixels();
  materials[7] = loadImage("mat08.png");
  materials[7].loadPixels();
  materials[8] = loadImage("mat09.png");
  materials[8].loadPixels();
  materials[9] = loadImage("mat10.png");
  materials[9].loadPixels();
  materials[10] = loadImage("mat11.png");
  materials[10].loadPixels();
  materials[11] = loadImage("mat12.png");
  materials[11].loadPixels();
  materials[12] = loadImage("mat13.png");
  materials[12].loadPixels();
  materials[13] = loadImage("mat14.png");
  materials[13].loadPixels();
  materials[14] = loadImage("mat15.png");
  materials[14].loadPixels();
  materials[15] = loadImage("mat16.png");
  materials[15].loadPixels();
  materials[16] = loadImage("mat17.png");
  materials[16].loadPixels();
  materials[17] = loadImage("mat18.png");
  materials[17].loadPixels();
  materials[18] = loadImage("mat19.png");
  materials[18].loadPixels();
  materials[19] = loadImage("mat20.png");
  materials[19].loadPixels();
  materials[20] = loadImage("mat21.png");
  materials[20].loadPixels();
  materials[21] = loadImage("mylar.png");
  materials[21].loadPixels();
}

public void loadBackgrounds(){
  kiteBackground[0] =  loadImage("whitebg.png");
  kiteBackground[1] =  loadImage("clearsky.png");
  kiteBackground[2] =  loadImage("lightclouds.png");
  kiteBackground[3] =  loadImage("darkclouds.png");
  kiteBackground[4] =  loadImage("sunset.png");
  activeBackground = 3;
}

public void setup(){
  size(620,630);

  guiBase = loadImage("GUI4Colorizer.png");
  guiY = kiteHeight+10;

  // Kite structures
  kiteTemplate = loadImage("kiteTemplate.png"); // Must be kiteWidth x kiteHeight pixels
  kiteShadow = loadImage("kiteShadow.png"); // Must be kiteWidth x kiteHeight pixels
  loadBackgrounds();
  loadMaterials();
  initMappings();
  kiteTemplate_create();
  kiteEdges_create();
    
  for(int i=0; i<numberMaterials; i++){
    materialButtons[i] = new ButtonI( 30+(i%6)*38, guiY+60+(int)(i/6)*38, 30, 30, materials[i]);
  }

  for(int i=0; i<5; i++){
    backgroundButtons[i] = new ButtonI( 300, guiY+60+i*38, 60, 30, kiteBackground[i]);
  }

  seamsButtons[0] = new ButtonC( 442, guiY+74, 15, 15, color(0,255,0,0));
  seamsButtons[1] = new ButtonC( 497, guiY+74, 15, 15, color(0,255,0,255));
  seamsButtons[2] = new ButtonC( 550, guiY+74, 15, 15, color(0,255,0,0));

  resetButton = new ButtonC( 497, guiY+280, 15, 15, color(255,0,0,255));
  effect3DButton = new ButtonC( 497, guiY+213, 15, 15, color(0,0,255,255));

  saveButton = new ButtonC( 497, guiY+246, 15, 15, color(255,0,255,255));

}
//** end of Setup & Initialization
//*****************************************************


//*****************************************************
//** Kite Templates and Visualization

public boolean isBorder(int k){
  boolean border = false;
  if(k > kiteWidth+1){
    border = border || ((kiteTemplate.pixels[k] != kiteTemplate.pixels[k-kiteWidth-1]));
    border = border || ((kiteTemplate.pixels[k] != kiteTemplate.pixels[k-kiteWidth]));
    border = border || ((kiteTemplate.pixels[k] != kiteTemplate.pixels[k-kiteWidth+1]));
  }
  if(k > 0) border = border || ((kiteTemplate.pixels[k] != kiteTemplate.pixels[k-1]));
  if(k < kiteWidth*kiteHeight-1) border = border || ((kiteTemplate.pixels[k] != kiteTemplate.pixels[k+1]));
  if(k < kiteWidth*(kiteHeight-1)-1){
    border = border || ((kiteTemplate.pixels[k] != kiteTemplate.pixels[kiteWidth+k-1]));
    border = border || ((kiteTemplate.pixels[k] != kiteTemplate.pixels[kiteWidth+k]));
    border = border || ((kiteTemplate.pixels[k] != kiteTemplate.pixels[kiteWidth+k+1]));
  }
  return border;
}

public void kiteEdges_create(){
  int i;
  
  kiteEdges = createImage(kiteWidth, kiteHeight, ARGB);
  kiteEdges.loadPixels();
  kiteTemplate.loadPixels();
  for(i=0; i < kiteWidth*kiteHeight; i++){
    if(isBorder(i)){
      kiteEdges.pixels[i] = color(0, 0, 0, 255);
    }
    else {
      kiteEdges.pixels[i] = color(0, 0, 0, 0);
    }
  }
  kiteEdges.updatePixels();
  kiteTemplate.updatePixels();
}

public void kiteTemplate_update(){
  int i,grey,mapping;
  float h, s, b, a, brght;
  
  colorMode(HSB);
  kiteTemplate.loadPixels();
  kiteShadow.loadPixels();
  kiteResult.loadPixels();
  for(i=0; i < kiteWidth*kiteHeight; i++){
    grey = (int)red(kiteTemplate.pixels[i]);
    if(grey==255) kiteResult.pixels[i] = color(0, 0, 0, 0);
    else{
      if(grey==0) kiteResult.pixels[i] = color(0, 0, 0, 255);
      else{
        mapping = mappingGreys2Materials[grey];
        kiteResult.pixels[i] = materials[mapping].pixels[i];
      }
    }
    if(effect3D){
      brght = (brightness(kiteShadow.pixels[i])+127.0f)/255.0f;
      if(brght>1.0f) brght = 1.0f;
      h = hue(kiteResult.pixels[i]);
      s = saturation(kiteResult.pixels[i]);
      b = brightness(kiteResult.pixels[i]);
      a = alpha(kiteResult.pixels[i]);
      kiteResult.pixels[i] = color(h, s, b*brght, a);
    }
  }
  kiteResult.updatePixels();
  kiteTemplate.updatePixels();
  kiteShadow.updatePixels();
  colorMode(RGB);
}

public void kiteTemplate_create(){  
  kiteResult = createImage(kiteWidth, kiteHeight, ARGB);
  kiteTemplate_update();
}
//** end of Kite Templates and Visualization
//*****************************************************

//*****************************************************
//** Draw
public void draw(){
  background(80);
  gui();

  kiteTemplate_update();
  noTint();  
  image(kiteBackground[activeBackground], 10,10);
  tint(255, 230);
  image(kiteResult, 10, 10);
  switch(seamsState){
    case 0: break;
    case 1: tint(255, 20);
            image(kiteEdges, 10, 10);
            break;
    case 2: tint(255, 200);
            image(kiteEdges, 10, 10);
            break;
    default: break;
  }
  noTint();  
}
//** end of Draw
//*****************************************************


  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "KiteColoriser_v1_0" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
